
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>


enum {
//	_xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 4_000_000,
	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,

	PINS = 20|7<<6,
	GROUP = 2,
	CLK_PIN = 15,

	TX_REGD = 1,
	CLK_REGD = 0,
	CPOL = 1,
	CPHA = 0,
	CLK_DIV = 2,

	M_NCO = 0x8000_0000UL / CLK_DIV + (0x8000_0000UL % CLK_DIV > 0UL ? 1UL : 0UL),    // round up
	M_LEADIN = X_IMM_32X1_1DAC1 | 7 + CLK_DIV + CLK_REGD - TX_REGD - (1 ^ CPHA) * (CLK_DIV>>1),

};



static uint32_t  pinout8clk_dma( unsigned pgroup, unsigned clkpin, uint8_t *haddr, unsigned len )
{
	uint32_t  m_tx = X_RFBYTE_8P_1DAC8 | pgroup<<20 | X_PINS_ON | len;  // byte wide transfers from hubRAM, via the FIFO
	uint32_t  datp = pgroup<<3 | 7<<6;
	uint32_t  ticks = _cnt();

	__asm volatile {  // no optimising and enforces Fcache use - Needed to free up the FIFO
		rdfast	#0, haddr    // setup the FIFO to read from hubRAM
		setxfrq	m_nco1    // set sysclock/1 for lead-in timing

		xinit	leadin, #0    // lead-in timing, at sysclock/1
		setq	m_nco2    // streamer transfer rate
		xcont	m_tx, #0    // setup buffered command for tx data
		drvl	datp    // active data pins for tx data
		dirh	clkpin    // start smartpin internally cycling, at SPI clock rate
		wypin	len, clkpin    // produce clocks, starting on second internal cycle

		// twiddle thumbs while DMA happens ...

		waitxfi   // wait for transfer to complete before returning back to hubexec, hubexec uses the FIFO
		dirl	clkpin    // reset smartpin
	_ret_	dirl	datp    // tristate the data bus upon completion

leadin		long	M_LEADIN
m_nco2		long	M_NCO
m_nco1		long	0x8000_0000
	}
	return  _cnt() - ticks;
}



static uint32_t  pinout8_dma( unsigned pgroup, uint8_t *haddr, unsigned len )
{
	uint32_t  m_nco = M_NCO;
	uint32_t  m_tx = X_RFBYTE_8P_1DAC8 | pgroup<<20 | X_PINS_ON | len;  // byte wide transfers from hubRAM, via the FIFO
	uint32_t  ticks = _cnt();
	
	__asm volatile {  // no optimising and enforces Fcache use - Needed to free up the FIFO
		rdfast	#0, haddr    // setup the FIFO to read from hubRAM
		setxfrq	m_nco    // set the transfer rate
		xinit	m_tx, #0    // do it!
		waitxfi   // wait for transfer to complete before returning back to hubexec, hubexec uses the FIFO
	}
	return  _cnt() - ticks;
}


static uint32_t  pinout8_grp4( unsigned pgroup, uint8_t *haddr, unsigned len )
{
	uint32_t  val, ticks = _cnt();

	__asm volatile {  // no optimising and enforces Fcache use - Needed to free up the FIFO
		rdfast	#0, haddr    // setup the FIFO to read from hubRAM
		rep	@.rend, len

		rfbyte	val    // read prefeched FIFO, intervals of 6 ticks
		altsb	pgroup, #0x1fc    //  SETBYTE indirection prefixer, OUTA base cog register
		setbyte	val     // "val" is indexed into 64-bit OUTA/OUTB register pair
.rend
	}
	return  _cnt() - ticks;
}


static uint32_t  pinout8_grp3( unsigned pgroup, uint8_t *haddr, unsigned len )
{
	uint32_t  val, ticks = _cnt();

	__asm volatile {  // no optimising and enforces Fcache use - To use cogexec branching speed in tight loop
		rep	@.rend, len    // setup a REP block to eliminate 4-tick branch from the loop

		rdbyte	val, haddr    // access hubRAM, intervals of 16 ticks (17 every four bytes)
		add	haddr, #1
		altsb	pgroup, #0x1fc    //  SETBYTE indirection prefixer, OUTA base cog register
		setbyte	val     // "val" is indexed into 64-bit OUTA/OUTB register pair
.rend
	}
	return  _cnt() - ticks;
}


static uint32_t  pinout8_grp2( unsigned pgroup, uint8_t *haddr, unsigned len )
{
	uint32_t  val, ticks = _cnt();

	__asm volatile {  // no optimising and enforces Fcache use - To use cogexec branching speed in tight loop
.loop
		rdbyte	val, haddr    // access hubRAM, intervals of 24 ticks (25 every four bytes)
		add	haddr, #1
		altsb	pgroup, #0x1fc    //  SETBYTE indirection prefixer, OUTA base cog register
		setbyte	val     // "val" is indexed into 64-bit OUTA/OUTB register pair
		djnz	len, #.loop
	}
	return  _cnt() - ticks;
}


static uint32_t  pinout8_grp1( unsigned pgroup, uint8_t *haddr, unsigned len )
{
	uint32_t  val, ticks = _cnt();

	__asm {    // allowed to be optimised and hubexec inlined
.loop
		rdbyte	val, haddr    // access hubRAM
		add	haddr, #1
		altsb	pgroup, #0x1fc    //  SETBYTE indirection prefixer, OUTA base cog register
		setbyte	val     // "val" is indexed into 64-bit OUTA/OUTB register pair
		djnz	len, #.loop
	}
	return  _cnt() - ticks;
}


static uint32_t  pinout8_grp( unsigned pgroup, uint8_t *haddr, unsigned len )
{
	uint8_t  val;
	uint32_t  i, ticks = _cnt();

	for( i=0; i<len; i++ )
	{
		val = *(haddr++);

		__asm {    // allowed to be optimised and hubexec inlined
		altsb	pgroup, #0x1fc    //  SETBYTE indirection prefixer, OUTA base cog register
		setbyte	val     // "val" is indexed into 64-bit OUTA/OUTB register pair
		}
	}

	return  _cnt() - ticks;
}


static uint32_t  pinout8_bb( unsigned pins, uint8_t *haddr, unsigned len )
{
	uint8_t  val;
	uint32_t  i, ticks = _cnt();

	for( i=0; i<len; i++ )  {
		val = *(haddr++);
		_pinwrite( pins, val );
	}

	return  _cnt() - ticks;
}


static uint32_t  filldat( uint8_t *haddr, unsigned len )
{
	uint32_t  i, ticks = _cnt();

	for( i = 1; i <= len; i++ )  {
//	for( i = len; i; i-- )  {
		*(haddr++) = (uint8_t)i;
	}

	return  _cnt() - ticks;
}



void  main( void )
{
	uint8_t  buff[1000];

	_waitms( 200 );
	printf( "\n   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode() );

	printf( "\narray size is %d bytes.  filldat() in %d ticks.\n", sizeof(buff), filldat( buff, sizeof(buff) ) );

	_pinf( GROUP<<3 | 7<<6 );
	_wrpin( GROUP<<3 | 7<<6, TX_REGD ? P_SYNC_IO : 0 );  // at sysclock/1, data pins should be registered to minimise relative skew

	_pinf( CLK_PIN );
	_wrpin( CLK_PIN, P_PULSE | P_OE | (CLK_REGD ? P_SYNC_IO : 0) | (CPOL ? P_INVERT_OUTPUT : 0) );    // clock pins
	_wxpin( CLK_PIN, CLK_DIV | (CLK_DIV>>1 + (CPOL & CLK_DIV))<<16 );    // set period and duty

	for(;;) // ever
	{
//		_pinl( PINS );
//		printf( "pinout8_bb() in %d ticks.\n", pinout8_bb( PINS, buff, sizeof(buff) ) );
//		_pinf( PINS );

//		printf( "pinout8_grp() in %d ticks.\n", pinout8_grp( GROUP, buff, sizeof(buff) ) );
//		printf( "pinout8_grp1() in %d ticks.\n", pinout8_grp1( GROUP, buff, sizeof(buff) ) );
//		printf( "pinout8_grp2() in %d ticks.\n", pinout8_grp2( GROUP, buff, sizeof(buff) ) );
//		printf( "pinout8_grp3() in %d ticks.\n", pinout8_grp3( GROUP, buff, sizeof(buff) ) );
//		printf( "pinout8_grp4() in %d ticks.\n", pinout8_grp4( GROUP, buff, sizeof(buff) ) );
//		printf( "pinout8_dma() in %d ticks.\n", pinout8_dma( GROUP, buff, sizeof(buff) ) );
		printf( "pinout8clk_dma() in %d ticks.\n\n", pinout8clk_dma( GROUP, CLK_PIN, buff, sizeof(buff) ) );

		_waitms( 800 );
	}
}
